﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.IO.Ports;
using ASCOM;
using ASCOM.Astrometry;
using ASCOM.Astrometry.AstroUtils;
using ASCOM.Utilities;

namespace ASCOM.AstroRoof
{
    class ArduinoSerial : SerialPort 
    {
        ASCOM.Utilities.Util HC = new ASCOM.Utilities.Util(); // Helper class
        public Stack CommandQueue = new Stack(); // Our command received stack

        public delegate void CommandQueueReadyEventHandler(object sender, EventArgs e); // Our Process stack callback
        public event CommandQueueReadyEventHandler CommandQueueReady;

        // Serial commands look like ": M 100 #"
        public struct SerialCommand
        {
            public static string GetStatus = "Status";
            public static string OpenShutter = "Open";
            public static string CloseShutter = "Close";

        }

        public ArduinoSerial(StopBits stopBits, int baud, bool autostart, string port)
        {
            this.Parity = Parity.None;
            this.PortName = port;
            this.StopBits = stopBits;
            this.BaudRate = baud;
            this.DataReceived += new SerialDataReceivedEventHandler(ArduinoSerial_DataReceived);

            if (autostart)
                this.Open();
        }

        public ArduinoSerial(string port) : this(StopBits.One, 9600, true, port) { }

        private void ArduinoSerial_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            CommandQueue.Push(this.ReadLine().Trim("\r".ToCharArray())); // Push latest command onto the stack
            CommandQueueReady(this, e);
        }

        public void SendCommand(string command)
        {
            this.Write(command);
        }

        public void ResetConnection()
        {
            this.Close();
            HC.WaitForMilliseconds(1000);
            this.Open();
            HC.WaitForMilliseconds(3000);
        }
    }
}
